/*
 * FILE: canarydump_x86_64.c
 *
 * Comment: A simple program which prints the value of the reference
 *          canary of a running process.
 *
 *          Get the canary of the init process
 *          $ sudo ./canarydump 1
 *
 *          Dump the canaries of all the process of the "www" account:
 *          $ for pid in `ps -u www-data | awk -F" " ' {if(NR>1)print $1;}'`; do
 *              sudo ./canarydump_x86_64 $pid;
 *            done
 *
 * Author: Hector Marco <hecmargi@upv.es>
 *         Ismael Ripoll <iripoll@disca.upv.es>
 *
 * Date:   February 2014
 *
 * Web:    http://renewssp.com    and    http://cybersecurity.upv.es
 * 
 * (C) Copyright, 2014  Hector Marco and Ismael Ripoll
 * 
 * This Software is provided to you solely for testing and
 * evaluation. You may not use it for any other purpose.
 *
 * The software is provided "as is", without warranty of any kind,
 * express or implied, including but not limited to the warranties of
 * merchantability, fitness for a particular purpose and
 * noninfringement. In no event shall the authors or copyright holders
 * be liable for any claim, damages or other liability, whether in an
 * action of contract, tort or otherwise, arising from, out of or in
 * connection with the software or the use or other dealings in the
 * software.
 */ 




#include <stdio.h>
#include <sys/ptrace.h>
#include <asm/ptrace.h>
#include <stdlib.h>
#include <sys/types.h>
#include <unistd.h>
#include <string.h>
#include <sys/user.h>
#include <linux/limits.h>
#include <sys/wait.h>
#include <stdarg.h>

#include <sys/types.h>
#include <sys/stat.h>
#include <fcntl.h>

#include <elf.h>
#include <sys/mman.h>

#include <sys/reg.h>

#ifdef __i386__
	#Error: This program only works in x86_64
#endif


/* Simple programs prefer to use global vars ;-) . */

pid_t pid;
char cmdline[PATH_MAX+100];
struct user_regs_struct regs;


#define REFERENCE_CANARY_OFFSET    (0x28)

void panic(char *format, ...){
        va_list argptr;

        va_start(argptr, format);
        fprintf(stderr, "Error: ");
        vfprintf(stderr, format, argptr);
	fprintf(stderr, "\n");
        va_end(argptr);
	exit(-1);
}

void attach_process(void){
	int status;

	if (ptrace(PTRACE_ATTACH, pid, NULL, NULL) == -1) {
	        panic("Attaching process [%d].\n", pid);
	}

	waitpid(pid, &status, 0x0);
}

void cont_process(void){
	if (ptrace(PTRACE_CONT, pid, NULL, NULL) == -1) { 
	        panic("Process [%d] can not be continued. Sorry.\n", pid);
	}
}

void dettach_process(void){
	int status;

	if (ptrace(PTRACE_DETACH, pid, NULL, NULL) == -1) {
	        panic("Detching process [%d].\n"), pid;
	}

	waitpid(pid, &status, 0x0);
}

void getRegisters(void){
        if (ptrace(PTRACE_GETREGS, pid, NULL, (void *)&regs) == -1) {
	        panic("Reading Registers from process [%d].\n", pid);
	}
}

unsigned long read_data_from_addr(unsigned long addr){

        return ptrace(PTRACE_PEEKDATA, pid, (void *)addr, NULL);

}

void get_cmdline(void){
	int fd;
	char pathToProc[128];
	if (snprintf(pathToProc, 128, "/proc/%d/cmdline", pid) > 22) {
	         panic("\"/proc/%d/cmdline\" entry too long.", pid);
	}
	fd = open(pathToProc, O_RDONLY);
	if (fd < 0){
	         panic("Can not open \"/proc/%d/cmdline\".", pid);
	}
	if (read(fd, cmdline, PATH_MAX+100) >= PATH_MAX) {
	         panic("Command line too long for pid [%d].", pid);
	}
	close(fd);
}

int main(int argc, const char *argv[]){
	unsigned long c_addr, referenceCanary;

	if ( geteuid() != 0 ) {
	        panic("You need root privileges to inspect other processes.");
	}


	if ( argv[1] == NULL ) {
	        panic("Please provide a PID.\n");
	}
	pid = (pid_t) atoi(argv[1]);

	if ( pid < 0 ) {
	        panic("Incorrect pid [%s].", argv[1]);
	}

	get_cmdline();

	attach_process();

	/* Get current registers */
	getRegisters();

	c_addr = regs.fs_base + REFERENCE_CANARY_OFFSET;

	referenceCanary = read_data_from_addr(c_addr);

	dettach_process();

	printf("Canary: [0x%016lx] Pid [%5d] %s\n", referenceCanary, pid, cmdline);

	return 0;
}
